/*
 *****************************************************
 *
 *  SaVi by Robert Thurman (thurman@geom.umn.edu) and
 *          Patrick Worfolk (worfolk@alum.mit.edu).
 *
 *  Copyright (c) 1997 by The Geometry Center.
 *  This file is part of SaVi.  SaVi is free software;
 *  you can redistribute it and/or modify it only under
 *  the terms given in the file COPYRIGHT which you should
 *  have received along with this file.  SaVi may be
 *  obtained by emailing savi@geom.umn.edu.
 *
 *****************************************************
 *
 *
 * coverage.c
 *
 * $Id: coverage.c,v 1.2 2002/06/23 17:26:28 lloydwood Exp $
 */
#include <stdlib.h>

#include "sats.h"
#include "constants.h"
#include "globals.h"
#include "coverage_vis.h"
#include "stats_utils.h"
#include "savi.h"

static double START_TIME;

static int coverage_flag = FALSE;
static int tracks_flag = FALSE;
int no_access_flag = FALSE;
int print_access_flag = FALSE;
static int project_tracks_flag = FALSE;
static int tracks_computed_flag = FALSE;
static int projection = SINUSOIDAL;
static int coverage_type = MASK_ELEVATION;
static int coverage_computed_flag = FALSE;

/* if c_grid is NULL we are about to draw first image */
static grid *c_grid = NULL;

static int N_OBSERVATIONS;

/* delta t (in seconds) for projecting ground tracks */
#define PROJECT_TRACKS_DT 5.0

/* Indices to be inserted in the foreground array for
 * projected ground tracks, and dynamically drawn ground tracks, resp. */

#define PROJECT_TRACKS_CI 2
#define GROUND_TRACKS_CI 3

/*
 * coverage_on_cmd -
 * Set flags to indicate that coverage should be actually computed and
 * displayed.  Call coverage_compute and coverage_display to get the ball
 * rolling. (These are repeatedly called during the sats_update cycle.)
 */
char *
coverage_on_cmd(int argc, char *argv[])
{
  Constellation *pconstellation = get_constellation();
  CentralBody *pcb = pconstellation->pcb;
  Satellite_list sl = pconstellation->satellites;

  /* do memory allocations and initialize display first time only */
  if (!c_grid) {
    c_grid = create_grid(IMAGE_HEIGHT, IMAGE_WIDTH);
    image_init(c_grid);
  }

  /* set variables so that coverage will be computed and displayed */
  coverage_flag = TRUE;
  coverage_computed_flag = FALSE;

  init_coverage_time();		/* reset average coverage info */
  coverage_compute(sl, FALSE, pcb);	/* compute the current coverage */
  coverage_display(sl, FALSE, pcb);	/* display the coverage map */

  return EMPTY_str;
}

char *
coverage_off_cmd(int argc, char *argv[])
{

  if (c_grid) {
    clear_intensity(c_grid);
    /*
     * to freeze coverage at a point in time (like interval
     * delta) do reset_foreground(projection, c_grid); instead
     */
    update_display(projection, c_grid, coverage_flag, tracks_flag, project_tracks_flag); /* used TRUE instead of coverage_flag lloyd */
  }
  coverage_flag = FALSE;

  return EMPTY_str;
}

char *
no_access_on_cmd(int argc, char *argv[])
{
  no_access_flag = TRUE;

  return EMPTY_str;
}

char *
no_access_off_cmd(int argc, char *argv[])
{
  no_access_flag = FALSE;

  return EMPTY_str;
}

char *
print_access_on_cmd(int argc, char *argv[])
{
  print_access_flag = TRUE;

  return EMPTY_str;
}

char *
print_access_off_cmd(int argc, char *argv[])
{
  print_access_flag = FALSE;

  return EMPTY_str;
}

char *
tracks_on_cmd(int argc, char *argv[])
{
  Constellation *pconstellation = get_constellation();
  Satellite_list sl = pconstellation->satellites;
  CentralBody *pcb = pconstellation->pcb;

  /* do memory allocations and initialize display first time only */
  if (!c_grid) {
    c_grid = create_grid(IMAGE_HEIGHT, IMAGE_WIDTH);
    image_init(c_grid);
  }

  /* set variable so tracks will be plotted */
  tracks_flag = TRUE;
  tracks_computed_flag = FALSE;

  tracks_compute(sl, pcb);	/* compute current tracks */
  coverage_display(sl, FALSE, pcb);	/* display the coverage map */

  return EMPTY_str;
}

char *
tracks_off_cmd(int argc, char *argv[])
{
  /* reset foreground map */
  if (c_grid) {
    reset_foreground(projection, c_grid);
    update_display(projection, c_grid, coverage_flag,
		   tracks_flag, project_tracks_flag);
  }

  tracks_flag = FALSE;

  return EMPTY_str;
}

/*
 * project_tracks_on_cmd -
 * Draw ground tracks forwards in time for all satellites.
 */
char *
project_tracks_on_cmd(int argc, char *argv[])
{
  Constellation *pconstellation = get_constellation();
  Satellite_list Sl = pconstellation->satellites;
  Satellite_list sl = Sl;
  CentralBody *pcb = pconstellation->pcb;

  double c_time = current_time();
  double f_time = c_time + tracks_interval * 60 * 60;	/* convert the tracks */
  double t, dt;			/* interval from hours */
  Satellite s;			/* to seconds. */
  int npts;

  if (tracks_interval >= 0)
    dt = PROJECT_TRACKS_DT;
  else
    dt = -PROJECT_TRACKS_DT;

  while (sl) {
    s = satellite_create(0, &(sl->s->oe), pcb);
    npts = (int) ((f_time - c_time) / dt);
    t = c_time;
    while ((npts--) >= 0) {
      satellite_compute_position(s, t, pcb);
      track_to_foreground(s, projection, c_grid, PROJECT_TRACKS_CI, pcb);
      t += dt;
    }
    satellite_destroy(s);
    sl = sl->next;
  }
  project_tracks_flag = TRUE;
  coverage_display(Sl, TRUE, pcb);	/* force the display */

  return EMPTY_str;
}

/*
 * project_tracks_off_cmd - Erase all ground tracks.
 */
char *
project_tracks_off_cmd(int argc, char *argv[])
{
  if (c_grid) {
    reset_foreground(projection, c_grid);
    update_display(projection, c_grid, coverage_flag,
		   tracks_flag, project_tracks_flag);
  }

  project_tracks_flag = FALSE;

  return EMPTY_str;
}

/*
 * project_tracks_reset
 */
void
project_tracks_reset()
{
  if (!project_tracks_flag)
    return;
  (void) project_tracks_off_cmd(0, NULL);	/* clear current ground tracks */
  (void) project_tracks_on_cmd(0, NULL);	/* and then re-draw. */
}

int
get_coverage_type()
{
  return coverage_type;
}

/*
 * fill out the pixel map for the Tk image
 */
void
coverage_compute(const Satellite_list SL, unsigned int force,
		 const CentralBody * pcb)
{
  if (!coverage_flag && !force)
    return;

  if (!c_grid)
    return;

  fill_grid(SL, projection, coverage_type, coverage_angle * DEG_TO_RAD, pcb,
	    c_grid);
  coverage_computed_flag = TRUE;
  update_coverage_time();
}

/*
 * tracks_compute - Plots point under satellites onto foreground image
 */
void
tracks_compute(const Satellite_list SL, const CentralBody * pcb)
{
  if (tracks_flag == FALSE)
    return;

  if (!c_grid)
    return;

  tracks_to_foreground(SL, projection, c_grid, GROUND_TRACKS_CI, pcb);
  tracks_computed_flag = TRUE;
}

/*
 * coverage_proj - sets type of projection
 */
char *
coverage_proj_cmd(int argc, char *argv[])
{
  Constellation *pconstellation = get_constellation();
  Satellite_list sl = pconstellation->satellites;
  CentralBody *pcb = pconstellation->pcb;
  int p = atoi(argv[2]);

  if (p < 0)
    p = 0;
  if (p >= NUM_PROJECTIONS)
    p = 0;
  projection = p;
  coverage_compute(sl, FALSE, pcb);	/* compute the current coverage */
  coverage_display(sl, FALSE, pcb);	/* display the coverage map */

  return EMPTY_str;
}

/*
 * coverage_angle_type -
 * sets type of angle for footprint computations - from earth or from satellite
 */
char *
coverage_angle_type_cmd(int argc, char *argv[])
{
  Constellation *pconstellation = get_constellation();
  Satellite_list sl = pconstellation->satellites;
  CentralBody *pcb = pconstellation->pcb;
  int t = atoi(argv[2]);

  coverage_type = t;
  coverage_compute(sl, TRUE, pcb);	/* compute the current coverage */
  coverage_display(sl, TRUE, pcb);	/* display the coverage map */

  /* rebuild the coverage cones, footprints, .. */
  for (t = 0; t < N_VIEW_MODULES; t++)
    if (view_modules[t].coverage_rebuild_fn)
      view_modules[t].coverage_rebuild_fn();

  return EMPTY_str;
}

/*
 * coverage_display - Display the coverage map.
 */
void
coverage_display(const Satellite_list SL, unsigned int force,
		 const CentralBody * pcb)
{
  if (!coverage_flag && !tracks_flag && !project_tracks_flag && !force)
    return;

  if (!c_grid)
    return;			/* image not yet created */

  if ((coverage_flag == TRUE) && (coverage_computed_flag == FALSE)) {
    coverage_compute(SL, force, pcb);
  }
  if ((tracks_flag == TRUE) && (tracks_computed_flag == FALSE)) {
    tracks_compute(SL, pcb);
  }
  update_display(projection, c_grid, coverage_flag,
		 tracks_flag, project_tracks_flag);
}

/*
 * init_coverage_time - Initialize variables for computing mean coverage.
 */
void
init_coverage_time()
{
  START_TIME = current_time();
  coverage_init_time_update(START_TIME);
  N_OBSERVATIONS = 1;
}

/*
 * update_coverage_time - Update coverage time, for use in computing mean coverage.
 */
void
update_coverage_time()
{
  coverage_time_update(current_time() - START_TIME);
}

/*
 * mean_coverage - Compute mean coverage over the current time interval.
 */
double
mean_coverage(int projection_type, grid * g)
{
  static double mean;

  mean =
    (mean * (N_OBSERVATIONS - 1) +
     total_coverage(projection_type, FALSE, g)) / N_OBSERVATIONS;
  N_OBSERVATIONS++;
  return (mean);
}
